"""Platform config client."""

from importlib.metadata import version

from .client import ConfigClient, ConfigClientBase
from .entities import (
    AMDGPU,
    GPU,
    ACMEEnvironment,
    AddNodePoolRequest,
    AMDGPUPreset,
    AppsConfig,
    ARecord,
    AWSCloudProvider,
    AWSCredentials,
    AWSStorage,
    AzureCloudProvider,
    AzureCredentials,
    AzureReplicationType,
    AzureStorage,
    AzureStorageTier,
    BucketsConfig,
    CloudProvider,
    CloudProviderOptions,
    CloudProviderType,
    Cluster,
    ClusterLocationType,
    ClusterStatus,
    CredentialsConfig,
    DisksConfig,
    DNSConfig,
    DockerRegistryConfig,
    EFSPerformanceMode,
    EFSThroughputMode,
    EMCECSCredentials,
    EnergyConfig,
    EnergySchedule,
    EnergySchedulePeriod,
    GoogleCloudProvider,
    GoogleFilestoreTier,
    GoogleStorage,
    GPUPreset,
    GrafanaCredentials,
    HelmRegistryConfig,
    IdleJobConfig,
    IngressConfig,
    IntelGPU,
    IntelGPUPreset,
    MetricsConfig,
    MinioCredentials,
    MonitoringConfig,
    NeuroAuthConfig,
    NodePool,
    NodePoolOptions,
    NodeRole,
    NotificationType,
    NvidiaGPU,
    NvidiaGPUPreset,
    OnPremCloudProvider,
    OpenStackCredentials,
    OrchestratorConfig,
    PatchClusterRequest,
    PatchNodePoolResourcesRequest,
    PatchNodePoolSizeRequest,
    PatchOrchestratorConfigRequest,
    PrometheusCredentials,
    PutNodePoolRequest,
    RegistryConfig,
    ResourcePoolType,
    ResourcePreset,
    Resources,
    SecretsConfig,
    SentryCredentials,
    Storage,
    StorageConfig,
    StorageInstance,
    TPUPreset,
    TPUResource,
    VCDCloudProvider,
    VCDCloudProviderOptions,
    VCDCredentials,
    VCDStorage,
    VolumeConfig,
)

__all__ = [
    "ConfigClient",
    "ConfigClientBase",
    "AMDGPU",
    "GPU",
    "ACMEEnvironment",
    "AddNodePoolRequest",
    "AMDGPUPreset",
    "AppsConfig",
    "ARecord",
    "AWSCloudProvider",
    "AWSCredentials",
    "AWSStorage",
    "AzureCloudProvider",
    "AzureCredentials",
    "AzureReplicationType",
    "AzureStorage",
    "AzureStorageTier",
    "BucketsConfig",
    "CloudProvider",
    "CloudProviderOptions",
    "CloudProviderType",
    "Cluster",
    "ClusterLocationType",
    "ClusterStatus",
    "CredentialsConfig",
    "DisksConfig",
    "DNSConfig",
    "DockerRegistryConfig",
    "EFSPerformanceMode",
    "EFSThroughputMode",
    "EMCECSCredentials",
    "EnergyConfig",
    "EnergySchedule",
    "EnergySchedulePeriod",
    "GoogleCloudProvider",
    "GoogleFilestoreTier",
    "GoogleStorage",
    "GPUPreset",
    "GrafanaCredentials",
    "HelmRegistryConfig",
    "IdleJobConfig",
    "IngressConfig",
    "IntelGPU",
    "IntelGPUPreset",
    "MetricsConfig",
    "MinioCredentials",
    "MonitoringConfig",
    "NeuroAuthConfig",
    "NodePool",
    "NodePoolOptions",
    "NodeRole",
    "NotificationType",
    "NvidiaGPU",
    "NvidiaGPUPreset",
    "OnPremCloudProvider",
    "OpenStackCredentials",
    "OrchestratorConfig",
    "PatchClusterRequest",
    "PatchNodePoolResourcesRequest",
    "PatchNodePoolSizeRequest",
    "PatchOrchestratorConfigRequest",
    "PrometheusCredentials",
    "PutNodePoolRequest",
    "RegistryConfig",
    "ResourcePoolType",
    "ResourcePreset",
    "Resources",
    "SecretsConfig",
    "SentryCredentials",
    "Storage",
    "StorageConfig",
    "StorageInstance",
    "TPUPreset",
    "TPUResource",
    "VCDCloudProvider",
    "VCDCloudProviderOptions",
    "VCDCredentials",
    "VCDStorage",
    "VolumeConfig",
]
__version__ = version(__name__)
